from typing import TYPE_CHECKING
from requests import Response, get


if TYPE_CHECKING:
    from opendart.models.common import ApiParams


def __check_status_code(response: Response):
    if response.status_code != 200:
        raise Exception(response.status_code, response.reason)


def __check_status(result: dict):
    if result['status'] != '000':
        raise Exception(result['status'], result['message'])


def fetch(url: str, params: 'ApiParams') -> Response:
    response = get(url, params)
    __check_status_code(response)
    return response


def parse_json(response: Response) -> dict:
    result: dict = response.json()
    __check_status(result)
    result.pop('status')
    result.pop('message')
    return result



from typing import TypedDict

from opendart.models.common import ApiParams


class Corporation(TypedDict):
    '''고유번호 DTO'''
    corp_code: str
    '''고유번호 - 공시대상회사의 고유번호(8자리)'''
    corp_name: str
    '''정식명칭 - 정식회사명칭'''
    stock_code: str
    '''종목코드 - 상장회사인 경우 주식의 종목코드(6자리)'''
    modify_date: str
    '''최종변경일자 - 기업개황정보 최종변경일자(YYYYMMDD)'''

class CorpOverview(TypedDict):
    '''기업개황 DTO'''
    corp_name: str
    '''정식명칭 - 정식회사명칭'''
    corp_name_eng: str
    '''영문명칭 - 영문정식회사명칭'''
    stock_name: str
    '''종목명(상장사) 또는 약식명칭(기타법인) - '''
    stock_code: str
    '''상장회사인 경우 주식의 종목코드(6자리)'''
    ceo_nm: str
    '''대표자명'''
    corp_cls: str
    '''법인구분 - Y(유가), K(코스닥), N(코넥스), E(기타)'''
    jurir_no: str
    '''법인등록번호'''
    bizr_no: str
    '''사업자등록번호'''
    adres: str
    '''주소'''
    hm_url: str
    '''홈페이지'''
    ir_url: str
    '''IR홈페이지'''
    phn_no: str
    '''전화번호'''
    fax_no: str
    '''팩스번호'''
    induty_code: str
    '''업종코드'''
    est_dt: str
    '''설립일(YYYYMMDD)'''
    acc_mt: str
    '''결산월(MM)'''


class CorpCodeParams(ApiParams):
    '''고유번호 Params'''

class CorpOverviewParams(ApiParams):
    '''기업개황 Params'''
    corp_code: str
    '''고유번호 - 공시대상회사의 고유번호(8자리)'''



from typing import Literal, TypedDict


RESPONSE_STATUS = Literal[ '000', '010', '011', '012', '013', '014', '020', '021', '100', '101', '800', '900', '901']
'''메시지 설명

000 : 정상
010 : 등록되지 않은 키입니다.
011 : 사용할 수 없는 키입니다. 오픈API에 등록되었으나, 일시적으로 사용 중지된 키를 통하여 검색하는 경우 발생합니다.
012 : 접근할 수 없는 IP입니다.
013 : 조회된 데이타가 없습니다.
014 : 파일이 존재하지 않습니다.
020 : 요청 제한을 초과하였습니다. 일반적으로는 10,000건 이상의 요청에 대하여 이 에러 메시지가 발생되나, 요청 제한이 다르게 설정된 경우에는 이에 준하여 발생됩니다.
021 : 조회 가능한 회사 개수가 초과하였습니다.(최대 100건)
100 : 필드의 부적절한 값입니다. 필드 설명에 없는 값을 사용한 경우에 발생하는 메시지입니다.
101 : 부적절한 접근입니다.
800 : 시스템 점검으로 인한 서비스가 중지 중입니다.
900 : 정의되지 않은 오류가 발생하였습니다.
901 : 사용자 계정의 개인정보 보유기간이 만료되어 사용할 수 없는 키입니다. 관리자 이메일(opendart@fss.or.kr)로 문의하시기 바랍니다.
'''

class ApiParams(TypedDict):
    crtfc_key: str
    '''API 인증키'''



from os import getenv

from dotenv import load_dotenv

load_dotenv(verbose=True)

DART_API_KEY: str = getenv('DART_API_KEY')
'''OpenDart API Key - .env로부터 API Key를 읽어옵니다.

example key = xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
'''

if not DART_API_KEY:
    raise Exception(".env에서 사용자의 API Key를 읽어오지 못했습니다.")



'''Async OpenDart APIs'''

from .public_notice import fetch_all_corp as fetch_all_corp
from .public_notice import fetch_all_listed_corp as fetch_all_listed_corp
from .public_notice import fetch_corp_overview_by_corp_code as fetch_corp_overview_by_corp_code

from .business_report import fetch_stock_total_quantity_status as fetch_stock_total_quantity_status
from .business_report import fetch_total_issued_stock_by_corp_code as fetch_total_issued_stock_by_corp_code



from typing import TYPE_CHECKING

from aiohttp import ClientSession, ClientResponse

from opendart.config import API_URLS

if TYPE_CHECKING:
    from opendart.models.common import ApiParams


def __check_status_code(response: ClientResponse):
    '''HTTP 통신 상태 확인'''
    if response.status != 200:
        raise Exception(response.status, response.reason)


def __check_status(result: dict):
    '''OPENDART로부터 받은 데이터가 정상인지 확인'''
    if result['status'] != '000':
        raise Exception(result['status'], result['message'])


def __is_binary(url: str) -> bool:
    return url in [
        API_URLS['corp_code']
    ]


async def __parse_xml_to_json(response: ClientResponse) -> dict:
    from io import BytesIO
    from json import loads, dumps
    from zipfile import ZipFile
    from xmltodict import parse

    _zip = ZipFile(BytesIO(await response.content.read()))
    _xml = parse(_zip.read('CORPCODE.xml').decode('utf-8'))
    return loads(dumps(_xml))


def __parse_json(result: dict) -> dict:
    __check_status(result)
    result.pop('status')
    result.pop('message')
    return result


async def fetch(session: ClientSession, url: str, params: 'ApiParams') -> dict:
    async with session.get(url, params=params) as response:

        __check_status_code(response)

        if __is_binary(url):
            return await __parse_xml_to_json(response)

        return __parse_json(await response.json())



import json
from typing import List, Union

def __append_json_extension(name: str) -> str:
    import re
    p = re.compile('\.json$')
    if not p.search(name):
        name += '.json'
    return name

def save_json_as_file(data: Union[dict, List[dict]], file_path: str) -> None:
    with open(__append_json_extension(file_path), 'w') as file:
        json.dump(data, file, ensure_ascii=False)

def read_json_by_file_path(file_path: str) -> Union[dict, List[dict]]:
    with open(__append_json_extension(file_path), 'r') as file:
        return json.load(file)



'''사업보고서 APIs

https://opendart.fss.or.kr/guide/main.do?apiGrpCd=DS002
'''

from typing import TYPE_CHECKING, List, Union

from opendart.api.common import fetch, parse_json
from opendart.config import API_URLS, DART_API_KEY

if TYPE_CHECKING:
    from opendart.models.business_report import StockQuantityStatus


def fetch_stock_total_quantity_status(
    corp_code: str,
    bsns_year: Union[str, int],
    reprt_code: Union[str, int] = 11011
) -> List['StockQuantityStatus']:
    '''주식의 총수 현황

    Parameters
    ----------
    corp_code: str
            고유번호(8자리)
    bsns_year: str or int
            사업연도(4자리) - 2015년도 이후 자료부터 제공
    reprt_code: str
            보고서 코드 - 1분기(11013), 반기(11012), 3분기(11014), 사업보고서(11011)
                default: 사업보고서(11011)

    - API URL: `https://opendart.fss.or.kr/api/stockTotqySttus.json`
    - API Params: `{
        crtfc_key: API_KEY,
        corp_code: 고유번호,
        bsns_year: 사업연도,
        reprt_code: 사업보고서(11011)
    }`
    '''

    url = API_URLS['stock_total_quantity']
    params = {
        'crtfc_key': DART_API_KEY,
        'corp_code': corp_code,
        'bsns_year': bsns_year,
        'reprt_code': reprt_code
    }
    return parse_json(fetch(url, params))['list']


def fetch_total_issued_stock_by_corp_code(
    corp_code: str,
    is_preferred_stock: bool = False
) -> int:
    '''발행 주식 총 수

    Dart에서 보통주인지 우선주인지에 따라 구분해야 하기 때문에
    우선주의 발행 주식 총 수를 구하려면 두 번째 인자에 명시해야 합니다.

    Parameters
    ----------
    corp_code: str
            고유번호(8자리)
    is_preferred_stock: bool
            우선주 여부
                default: False

    # TODO 사업계획서(11011) 기준이라 bsns_year 값을 작년으로 잡았는데, 이게 맞나?
    '''
    from datetime import date
    reports = fetch_stock_total_quantity_status(
        corp_code,
        date.today().year - 1
    )
    def compare(x: 'StockQuantityStatus') -> bool:
        return x['se'] == ('우선주' if is_preferred_stock else '보통주')
    total_issued_stock = list(filter(compare, reports)).pop()
    return int(total_issued_stock.get('distb_stock_co').replace(',',''))



from typing import TypedDict


class ApiUrls(TypedDict):
    corp_code: str
    '''고유번호'''
    corp_overview: str
    '''기업개황'''
    stock_total_quantity: str
    '''주식의 총수 현황'''

API_URLS: ApiUrls = {
    'corp_code': 'https://opendart.fss.or.kr/api/corpCode.xml',
    'corp_overview': 'https://opendart.fss.or.kr/api/company.json',
    'stock_total_quantity': 'https://opendart.fss.or.kr/api/stockTotqySttus.json',
}
'''API URL 목록'''



from setuptools import setup, find_packages

setup(
    name                          = 'OpenDart',
    version                       = '0.1.0',
    description                   = '전자공시 OPEN DART 시스템 API',
    long_description              = open('README.md').read(),
    long_description_content_type = 'text/markdown',
    author                        = 'jihogrammer',
    author_email                  = 'jihogrammer@gmail.com',
    license                       = 'MIT',
    packages                      = find_packages(),
    install_requires              = ['python-dotenv', 'requests', 'aiohttp', 'xmltodict'],
    classifiers                   = [
        # https://pypi.org/classifiers
        "Development Status :: 2 - Pre-Alpha",
        "Natural Language :: Korean",
        "Environment :: Web Environment",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Topic :: Office/Business",
    ],
)



'''공시정보 APIs

https://opendart.fss.or.kr/guide/main.do?apiGrpCd=DS001
'''

import json
from typing import TYPE_CHECKING, List


from opendart.api.common import fetch, parse_json
from opendart.config import API_URLS, DART_API_KEY

if TYPE_CHECKING:
    from requests import Response
    from opendart.models.public_notice import Corporation, CorpCodeParams, CorpOverview, CorpOverviewParams


def __parse_xml_to_json(response: 'Response') -> dict:
    from io import BytesIO
    from zipfile import ZipFile
    from xmltodict import parse

    _zip = ZipFile(BytesIO(response.content))
    _xml = parse(_zip.read('CORPCODE.xml').decode('utf-8'))
    return json.loads(json.dumps(_xml))


def fetch_all_corp() -> List['Corporation']:
    '''고유번호 조회

    - API URL: `https://opendart.fss.or.kr/api/corpCode.xml`
    - API Params: `{ crtfc_key: API_KEY }`

    공시대상회사 목록을 조회합니다.

    WARNING
    -------
    다른 API와는 다르게 xml 방식만 지원합니다.
    하지만 실제 반환값(출력포맷)은 Zip file이기 때문에,
    압축을 풀고 xml을 json으로 변환하는 작업이 수행됩니다.

    시간이 걸리는 API이므로 매번 호출하지 않고,
    조회 결과 리스트를 별도 파일로 저장하거나 DB와 연동하여 관리하는 것을 권장합니다.
    '''

    url = API_URLS['corp_code']
    params: 'CorpCodeParams' = { 'crtfc_key': DART_API_KEY }
    response = fetch(url, params)

    return __parse_xml_to_json(response)['result']['list']


def fetch_all_listed_corp() -> List['Corporation']:
    '''상장회사 목록 조회

    공시대상회사 목록 중 상장사 목록을 조회합니다.

    WARNING
    -------
    약 9만 개 공시대상회사 목록 중 상장회사 목록을 조회합니다.
    고유번호 조회 후 필터링하기 때문에 시간이 다소 걸릴 수 있습니다.
    따라서 매번 호출하지 않고 별도 파일로 저장하거나 DB를 연동해서 관리하는 것을 권장합니다.

    또한, 이 API는 Dart에서 직접 제공하지 않아 편의를 위해 만들었습니다.
    단순히 고유번호 조회 API에서 종목코드를 보유한 회사 목록으로 걸러냈습니다.
    회사 고유번호를 별도로 관리하고 있다면 직접 개발하는 것을 권장합니다.

    See also :func:`api.public_notice.fetch_all_corp()`
    '''
    return list(filter(lambda c: c['stock_code'], fetch_all_corp()))


def fetch_corp_overview_by_corp_code(corp_code: str) -> 'CorpOverview':
    '''기업개황 조회

    - API URL: `https://opendart.fss.or.kr/api/company.json`
    - API Params `{
        crtfc_key: API_KEY,
        corp_code: 고유번호
    }`

    고유번호를 통해 해당하는 기업의 개황정보를 조회합니다.
    '''
    url = API_URLS['corp_overview']
    params: 'CorpOverviewParams' = {
        'crtfc_key': DART_API_KEY,
        'corp_code': corp_code
    }
    return parse_json(fetch(url, params))



from typing import TypedDict


class StockQuantityStatus(TypedDict):
    rcept_no: str
    '''접수번호 - 접수번호(14자리)'''
    corp_cls: str
    '''법인구분 - Y(유가), K(코스닥), N(코넥스), E(기타)'''
    corp_code: str
    '''고유번호 - 공시대상회사의 고유번호(8자리)'''
    corp_name: str
    '''회사명 - 공시대상회사명'''
    se: str
    '''구분(증권의종류, 합계, 비고)'''
    isu_stock_totqy: str
    '''발행할 주식의 총수'''
    now_to_isu_stock_totqy: str
    '''현재까지 발행한 주식의 총수'''
    now_to_dcrs_stock_totqy: str
    '''현재까지 감소한 주식의 총수'''
    redc: str
    '''감자 - 현재까지 감소한 주식의 총수'''
    profit_incnr: str
    '''이익소각 - 현재까지 감소한 주식의 총수'''
    rdmstk_repy: str
    '''상환주식의 상환 - 현재까지 감소한 주식의 총수'''
    etc: str
    '''기타 - 현재까지 감소한 주식의 총수'''
    istc_totqy: str
    '''발행주식의 총수'''
    tesstk_co: str
    '''자기주식수'''
    distb_stock_co: str
    '''유통주식수'''
